"""
Define different custom number types.

The same functionality can be implemented in different ways,

to provide more options, and open the more various and wider possibilities.
"""

from crimson.types_beta.addon.intelli_type import IntelliHolder, T
from crimson.types_beta.addon.annotated import AnnotatedType


class IntRange(AnnotatedType, int, IntelliHolder[T]):
    """
    Int Type, with additional information injection capability.

    Example:
        - [link](https://github.com/crimson206/types/blob/main/example/control/number.ipynb) to example notebook
    """


class FloatRange(
    AnnotatedType,
    float,
    IntelliHolder[T],
):
    """
    Float Type, with additional information injection capability.

    Example:
        - [link](https://github.com/crimson206/types/blob/main/example/control/highlighted/number.ipynb.md) to example notebook
    """



from torch import Tensor


class Query(Tensor):
    """
    Query tensor for attention mechanisms.

    Shape: (batch_size, n_head, seq_len_q, d_k)
    - n_head: Number of attention heads
    - seq_len_q: Query sequence length
    - d_k: Dimension of keys/queries (usually d_model // n_head)
    """


class Key(Tensor):
    """
    Key tensor for attention mechanisms.

    Shape: (batch_size, n_head, seq_len_k, d_k)
    - n_head: Number of attention heads
    - seq_len_k: Key sequence length
    - d_k: Dimension of keys/queries (usually d_model // n_head)
    """


class Value(Tensor):
    """
    Value tensor for attention mechanisms.

    Shape: (batch_size, n_head, seq_len_v, d_v)
    - n_head: Number of attention heads
    - seq_len_v: Value sequence length (usually same as seq_len_k)
    - d_v: Dimension of values (usually d_model // n_head)
    """


class Attn(Tensor):
    """
    Attention weights tensor.

    Shape: (batch_size, n_head, seq_len_q, seq_len_k)
    - n_head: Number of attention heads
    - seq_len_q: Query sequence length
    - seq_len_k: Key sequence length
    """


class QueriedValue(Tensor):
    """
    Output tensor after applying attention weights to values.

    Shape: (batch_size, n_head, seq_len_q, d_v)
    - n_head: Number of attention heads
    - seq_len_q: Query sequence length
    - d_v: Dimension of values
    """



from typing import Generic, TypeVar

T = TypeVar("T")


class IntelliHolder(Generic[T]):
    """
    It is Generic[T] holder.

    You can inherit to your custom type class, and \n
    when you hover on your function, the function will show you \n
            - not only the name of your custom type,
            - but also it's inner type.

    Example:
        - [link](https://github.com/crimson206/types/blob/main/example/addon/intelli_type.ipynb.md) to example
    """
